@@ -1,3 +1,5 @@
+# Copyright (c) Microsoft Corporation. All rights reserved.
+# Licensed under the MIT License.
 """Unittest helpers.
 
 All functions in this file should be considered private APIs,
@@ -13,9 +15,10 @@ import inspect
 import json
 import logging
 import os
-import queue
 import pathlib
 import platform
+import queue
+import re
 import shutil
 import socket
 import subprocess
@@ -25,7 +28,6 @@ import time
 import typing
 import unittest
 import uuid
-import re
 
 import grpc
 import requests
@@ -33,9 +35,8 @@ import requests
 from azure_functions_worker._thirdparty import aio_compat
 from . import dispatcher
 from . import protos
-from .utils.common import is_envvar_true
 from .constants import PYAZURE_WEBHOST_DEBUG
-
+from .utils.common import is_envvar_true
 
 PROJECT_ROOT = pathlib.Path(__file__).parent.parent
 TESTS_ROOT = PROJECT_ROOT / 'tests'
@@ -43,13 +44,14 @@ E2E_TESTS_FOLDER = pathlib.Path('endtoend')
 E2E_TESTS_ROOT = TESTS_ROOT / E2E_TESTS_FOLDER
 UNIT_TESTS_FOLDER = pathlib.Path('unittests')
 UNIT_TESTS_ROOT = TESTS_ROOT / UNIT_TESTS_FOLDER
-DEFAULT_WEBHOST_DLL_PATH = PROJECT_ROOT / 'build' / 'webhost' / \
-    'Microsoft.Azure.WebJobs.Script.WebHost.dll'
+WEBHOST_DLL = "Microsoft.Azure.WebJobs.Script.WebHost.dll"
+DEFAULT_WEBHOST_DLL_PATH = PROJECT_ROOT / 'build' / 'webhost' / WEBHOST_DLL
 EXTENSIONS_PATH = PROJECT_ROOT / 'build' / 'extensions' / 'bin'
 FUNCS_PATH = TESTS_ROOT / UNIT_TESTS_FOLDER / 'http_functions'
 WORKER_PATH = PROJECT_ROOT / 'python' / 'test'
 WORKER_CONFIG = PROJECT_ROOT / '.testconfig'
 ON_WINDOWS = platform.system() == 'Windows'
+LOCALHOST = "127.0.0.1"
 
 HOST_JSON_TEMPLATE = """\
 {
@@ -146,7 +148,7 @@ class WebHostTestCaseMeta(type(unittest.TestCase)):
                         # Trim off host output timestamps
                         host_output = getattr(self, 'host_out', '')
                         output_lines = host_output.splitlines()
-                        ts_re = r"^\[\d+\/\d+\/\d+ \d+\:\d+\:\d+ (A|P)M\]"
+                        ts_re = r"^\[\d+\/\d+\/\d+ \d+\:\d+\:\d+.*(A|P)*M*\]"
                         output = list(map(
                             lambda s: re.sub(ts_re, '', s).strip(),
                             output_lines))
@@ -169,6 +171,11 @@ class WebHostTestCase(unittest.TestCase, metaclass=WebHostTestCaseMeta):
     In addition to automatically starting up a WebHost instance,
     this test case class logs WebHost stdout/stderr in case
     a unit test fails.
+
+    You can write two sets of test - test_* and check_log_* tests.
+
+    test_ABC - Unittest
+    check_log_ABC - Check logs generated during the execution of test_ABC.
     """
     host_stdout_logger = logging.getLogger('webhosttests')
 
@@ -245,7 +252,7 @@ class _MockWebHostServicer(protos.FunctionRpcServicer):
                     f'unexpected {rtype!r} initial message from the worker')
 
             if client_response.start_stream.worker_id != self._host.worker_id:
-                raise AssertionError(f'worker_id mismatch')
+                raise AssertionError('worker_id mismatch')
 
         except Exception as ex:
             self._host._loop.call_soon_threadsafe(
@@ -316,7 +323,7 @@ class _MockWebHost:
         self._server = grpc.server(self._threadpool)
         self._servicer = _MockWebHostServicer(self)
         protos.add_FunctionRpcServicer_to_server(self._servicer, self._server)
-        self._port = self._server.add_insecure_port('127.0.0.1:0')
+        self._port = self._server.add_insecure_port(f'{LOCALHOST}:0')
 
         self._worker_id = self.make_id()
         self._request_id = self.make_id()
@@ -408,10 +415,10 @@ class _MockWebHost:
         self._in_queue.put_nowait((message, wait_for))
         return await self._out_aqueue.get()
 
-    async def _start(self):
+    async def start(self):
         self._server.start()
 
-    async def _close(self):
+    async def close(self):
         self._in_queue.put_nowait((_MockWebHostServicer._STOP, None))
         self._server.stop(1)
 
@@ -455,21 +462,20 @@ class _MockWebHostController:
         loop = aio_compat.get_running_loop()
         self._host = _MockWebHost(loop, self._scripts_dir)
 
-        await self._host._start()
+        await self._host.start()
 
-        self._worker = await dispatcher.Dispatcher.connect(
-            '127.0.0.1', self._host._port,
-            self._host.worker_id, self._host.request_id,
-            connect_timeout=5.0)
-
-        self._worker.load_bindings()
+        self._worker = await dispatcher. \
+            Dispatcher.connect(LOCALHOST, self._host._port,
+                               self._host.worker_id,
+                               self._host.request_id, connect_timeout=5.0)
 
         self._worker_task = loop.create_task(self._worker.dispatch_forever())
 
-        done, pending = await asyncio.wait(
-            [self._host._connected_fut, self._worker_task],
-            return_when=asyncio.FIRST_COMPLETED)
+        done, pending = await asyncio. \
+            wait([self._host._connected_fut, self._worker_task],
+                 return_when=asyncio.FIRST_COMPLETED)
 
+        # noinspection PyBroadException
         try:
             if self._worker_task in done:
                 self._worker_task.result()
@@ -478,7 +484,7 @@ class _MockWebHostController:
                 raise RuntimeError('could not start a worker thread')
         except Exception:
             try:
-                self._host._close()
+                await self._host.close()
                 self._worker.stop()
             finally:
                 raise
@@ -496,14 +502,14 @@ class _MockWebHostController:
         self._worker_task = None
         self._worker = None
 
-        await self._host._close()
+        await self._host.close()
         self._host = None
 
 
 def start_mockhost(*, script_root=FUNCS_PATH):
     tests_dir = TESTS_ROOT
     scripts_dir = tests_dir / script_root
-    if not scripts_dir.exists() or not scripts_dir.is_dir():
+    if not (scripts_dir.exists() and scripts_dir.is_dir()):
         raise RuntimeError(
             f'invalid script_root argument: '
             f'{scripts_dir} directory does not exist')
@@ -537,7 +543,7 @@ class _WebHostProxy:
 
 def _find_open_port():
     with socket.socket() as s:
-        s.bind(('127.0.0.1', 0))
+        s.bind((LOCALHOST, 0))
         s.listen(1)
         return s.getsockname()[1]
 
@@ -573,10 +579,10 @@ def popen_webhost(*, stdout, stderr, script_root=FUNCS_PATH, port=None):
         if not dll:
             dll = DEFAULT_WEBHOST_DLL_PATH
 
-            secrets = SECRETS_TEMPLATE
-
             os.makedirs(dll.parent / 'Secrets', exist_ok=True)
             with open(dll.parent / 'Secrets' / 'host.json', 'w') as f:
+                secrets = SECRETS_TEMPLATE
+
                 f.write(secrets)
 
         if dll and pathlib.Path(dll).exists():
@@ -584,33 +590,29 @@ def popen_webhost(*, stdout, stderr, script_root=FUNCS_PATH, port=None):
 
     if not hostexe_args:
         raise RuntimeError('\n'.join([
-            f'Unable to locate Azure Functions Host binary.',
-            f'Please do one of the following:',
-            f' * run the following command from the root folder of',
-            f'   the project:',
-            f'',
+            'Unable to locate Azure Functions Host binary.',
+            'Please do one of the following:',
+            ' * run the following command from the root folder of',
+            '   the project:',
+            '',
             f'       $ {sys.executable} setup.py webhost',
-            f'',
-            f' * or download or build the Azure Functions Host and'
-            f'   then write the full path to WebHost.dll'
-            f'   into the `PYAZURE_WEBHOST_DLL` environment variable.',
-            f'   Alternatively, you can create the',
+            '',
+            ' * or download or build the Azure Functions Host and'
+            '   then write the full path to WebHost.dll'
+            '   into the `PYAZURE_WEBHOST_DLL` environment variable.',
+            '   Alternatively, you can create the',
             f'   {WORKER_CONFIG.name} file in the root folder',
-            f'   of the project with the following structure:',
-            f'',
-            f'      [webhost]',
-            f'      dll = /path/Microsoft.Azure.WebJobs.Script.WebHost.dll',
-            f' * or download Azure Functions Core Tools binaries and',
-            f'   then write the full path to func.exe into the ',
-            f'   `CORE_TOOLS_PATH` envrionment variable.'
+            '   of the project with the following structure:',
+            '',
+            '      [webhost]',
+            '      dll = /path/Microsoft.Azure.WebJobs.Script.WebHost.dll',
+            ' * or download Azure Functions Core Tools binaries and',
+            '   then write the full path to func.exe into the ',
+            '   `CORE_TOOLS_PATH` envrionment variable.'
         ]))
 
     worker_path = os.environ.get('PYAZURE_WORKER_DIR')
-    if not worker_path:
-        worker_path = WORKER_PATH
-    else:
-        worker_path = pathlib.Path(worker_path)
-
+    worker_path = WORKER_PATH if not worker_path else pathlib.Path(worker_path)
     if not worker_path.exists():
         raise RuntimeError(f'Worker path {worker_path} does not exist')
 
@@ -641,6 +643,15 @@ def popen_webhost(*, stdout, stderr, script_root=FUNCS_PATH, port=None):
         if servicebus:
             extra_env['AzureWebJobsServiceBusConnectionString'] = servicebus
 
+        eventgrid_topic_uri = testconfig['azure'].get('eventgrid_topic_uri')
+        if eventgrid_topic_uri:
+            extra_env['AzureWebJobsEventGridTopicUri'] = eventgrid_topic_uri
+
+        eventgrid_topic_key = testconfig['azure'].get('eventgrid_topic_key')
+        if eventgrid_topic_key:
+            extra_env['AzureWebJobsEventGridConnectionKey'] = \
+                eventgrid_topic_key
+
     if port is not None:
         extra_env['ASPNETCORE_URLS'] = f'http://*:{port}'
 
@@ -656,11 +667,7 @@ def popen_webhost(*, stdout, stderr, script_root=FUNCS_PATH, port=None):
 
 
 def start_webhost(*, script_dir=None, stdout=None):
-    if script_dir:
-        script_root = TESTS_ROOT / script_dir
-    else:
-        script_root = FUNCS_PATH
-
+    script_root = TESTS_ROOT / script_dir if script_dir else FUNCS_PATH
     if stdout is None:
         if is_envvar_true(PYAZURE_WEBHOST_DEBUG):
             stdout = sys.stdout
@@ -670,24 +677,27 @@ def start_webhost(*, script_dir=None, stdout=None):
     port = _find_open_port()
     proc = popen_webhost(stdout=stdout, stderr=subprocess.STDOUT,
                          script_root=script_root, port=port)
+    time.sleep(10)  # Giving host some time to start fully.
 
-    addr = f'http://127.0.0.1:{port}'
-    for n in range(10):
+    addr = f'http://{LOCALHOST}:{port}'
+    for _ in range(10):
         try:
             r = requests.get(f'{addr}/api/ping',
                              params={'code': 'testFunctionKey'})
+            # Give the host a bit more time to settle
+            time.sleep(2)
+
             if 200 <= r.status_code < 300:
                 # Give the host a bit more time to settle
                 time.sleep(2)
                 break
         except requests.exceptions.ConnectionError:
             pass
-
-        time.sleep(0.5)
+        time.sleep(2)
     else:
         proc.terminate()
         try:
-            proc.wait(10)
+            proc.wait(20)
         except subprocess.TimeoutExpired:
             proc.kill()
         raise RuntimeError('could not start the webworker')
@@ -698,7 +708,7 @@ def start_webhost(*, script_dir=None, stdout=None):
 def create_dummy_dispatcher():
     dummy_event_loop = asyncio.new_event_loop()
     disp = dispatcher.Dispatcher(
-        dummy_event_loop, '127.0.0.1', 0,
+        dummy_event_loop, LOCALHOST, 0,
         'test_worker_id', 'test_request_id',
         1.0, 1000)
     dummy_event_loop.close()
@@ -726,7 +736,7 @@ def retryable_test(
     return decorate
 
 
-def _remove_path(path):
+def remove_path(path):
     if path.is_symlink():
         path.unlink()
     elif path.is_dir():
@@ -736,7 +746,7 @@ def _remove_path(path):
 
 
 def _symlink_dir(src, dst):
-    _remove_path(dst)
+    remove_path(dst)
 
     if ON_WINDOWS:
         shutil.copytree(str(src), str(dst))
@@ -749,8 +759,9 @@ def _setup_func_app(app_root):
     ping_func = app_root / 'ping'
     host_json = app_root / 'host.json'
 
-    with open(host_json, 'w') as f:
-        f.write(HOST_JSON_TEMPLATE)
+    if not os.path.isfile(host_json):
+        with open(host_json, 'w') as f:
+            f.write(HOST_JSON_TEMPLATE)
 
     _symlink_dir(TESTS_ROOT / 'common' / 'ping', ping_func)
     _symlink_dir(EXTENSIONS_PATH, extensions)
@@ -762,7 +773,7 @@ def _teardown_func_app(app_root):
     host_json = app_root / 'host.json'
 
     for path in (extensions, ping_func, host_json):
-        _remove_path(path)
+        remove_path(path)
 
 
 def _main():
